import { RawEdge, RawNode } from '@linkurious/ogma';
import csv from 'csv-parser';
import { v4 as uuidv4 } from 'uuid';
import * as fs from 'fs';
import {
  Movie,
  Person,
  Genre,
  DirectedRelationship,
  ActedInRelationship,
  HasGenreRelationship,
  MovieRow,
  NodeType,
  EdgeType
} from './types'; // Adjust the path as needed

type UUID = string;

export async function parse(csvFilePath: string): Promise<{
  movies: Movie[];
  people: Person[];
  genres: Genre[];
  directedRelationships: DirectedRelationship[];
  actedInRelationships: ActedInRelationship[];
  hasGenreRelationships: HasGenreRelationship[];
}> {
  const movies: Movie[] = [];
  const people: Person[] = [];
  const genres: Genre[] = [];
  const directedRelationships: DirectedRelationship[] = [];
  const actedInRelationships: ActedInRelationship[] = [];
  const hasGenreRelationships: HasGenreRelationship[] = [];

  const { movieMap, personMap, genreMap } = await prepareIds(csvFilePath);

  // Convert maps to arrays for insertion
  for (const [name, id] of personMap.entries()) {
    people.push({ id, name });
  }

  for (const [name, id] of genreMap.entries()) {
    genres.push({ id, name });
  }

  // Process CSV data for movies and relationships
  await new Promise<void>((resolve, reject) => {
    const moviesSet = new Set<string>();
    fs.createReadStream(csvFilePath)
      .pipe(csv())
      .on('data', (row: MovieRow) => {
        // Process Movie using pre-generated ID
        const movieId = movieMap.get(row.Title)!;
        if (moviesSet.has(movieId)) {
          console.log(`Duplicate movie found: ${row.Title}. Skipping.`);
          return;
        }
        moviesSet.add(movieId);
        const movie: Movie = {
          id: movieId,
          title: row.Title,
          description: row.Description,
          year: parseInt(String(row.Year)),
          runtime: parseInt(String(row['Runtime (Minutes)'])),
          votes: parseInt(String(row.Votes))
        };
        movies.push(movie);

        // Process Director(s) relationships
        const directors = row.Director.split(',').map(d => d.trim());
        directors.forEach(directorName => {
          const personId = personMap.get(directorName)!;

          // Create directed relationship
          directedRelationships.push({
            id: uuidv4(),
            personId: personId,
            movieId: movieId
          });
        });

        // Process Actor(s) relationships
        const actors = row.Actors.split(',').map(a => a.trim());
        actors.forEach(actorName => {
          const personId = personMap.get(actorName)!;

          // Create acted_in relationship
          actedInRelationships.push({
            id: uuidv4(),
            personId: personId,
            movieId: movieId
          });
        });

        // Process Genre(s) relationships
        const genreList = row.Genre.split(',').map(g => g.trim());
        genreList.forEach(genreName => {
          const genreId = genreMap.get(genreName)!;

          // Create has_genre relationship
          hasGenreRelationships.push({
            id: uuidv4(),
            movieId: movieId,
            genreId: genreId
          });
        });
      })
      .on('end', () => {
        console.log(`Processed data:`);
        console.log(`- ${movies.length} movies`);
        console.log(`- ${people.length} people`);
        console.log(`- ${genres.length} genres`);
        console.log(`- ${directedRelationships.length} directed relationships`);
        console.log(`- ${actedInRelationships.length} acted_in relationships`);
        console.log(
          `- ${hasGenreRelationships.length} has_genre relationships`
        );

        fs.writeFileSync(
          'drop.json',
          JSON.stringify(
            toGraph(
              movies,
              people,
              genres,
              directedRelationships,
              actedInRelationships,
              hasGenreRelationships
            ),
            undefined,
            0
          ),
          'utf-8'
        );

        resolve();
      })
      .on('error', error => {
        reject(error);
      });
  });

  return {
    movies,
    people,
    genres,
    directedRelationships,
    actedInRelationships,
    hasGenreRelationships
  };
}

async function prepareIds(csvFilePath: string) {
  console.log('Preparing consistent IDs for all entities...');
  // Store IDs consistently for relationships
  const movieMap = new Map<string, UUID>(); // Title -> ID mapping
  const personMap = new Map<string, UUID>(); // Person name -> ID mapping
  const genreMap = new Map<string, UUID>(); // Genre name -> ID mapping

  await new Promise<void>((resolve, reject) => {
    fs.createReadStream(csvFilePath!)
      .pipe(csv())
      .on('data', (row: MovieRow) => {
        // Generate consistent ID for movie
        const movieId = uuidv4();

        if (movieMap.has(row.Title)) {
          console.log(
            `Duplicate movie title found: ${row.Title}. Using existing ID.`
          );
        }
        movieMap.set(row.Title, movieId);

        // Generate consistent IDs for directors
        const directors = row.Director.split(',').map(d => d.trim());
        directors.forEach(directorName => {
          if (!personMap.has(directorName)) {
            personMap.set(directorName, uuidv4());
          }
        });

        // Generate consistent IDs for actors
        const actors = row.Actors.split(',').map(a => a.trim());
        actors.forEach(actorName => {
          if (!personMap.has(actorName)) {
            personMap.set(actorName, uuidv4());
          }
        });

        // Generate consistent IDs for genres
        const genreList = row.Genre.split(',').map(g => g.trim());
        genreList.forEach(genreName => {
          if (!genreMap.has(genreName)) {
            genreMap.set(genreName, uuidv4());
          }
        });
      })
      .on('end', () => {
        console.log(`Prepared IDs for:`);
        console.log(`- ${movieMap.size} movies`);
        console.log(`- ${personMap.size} people (actors & directors)`);
        console.log(`- ${genreMap.size} genres`);
        resolve();
      })
      .on('error', error => {
        reject(error);
      });
  });

  return {
    movieMap,
    personMap,
    genreMap
  };
}

function toGraph(
  movies: Movie[],
  people: Person[],
  genres: Genre[],
  directedRelationships: DirectedRelationship[],
  actedInRelationships: ActedInRelationship[],
  hasGenreRelationships: HasGenreRelationship[]
) {
  const nodes: RawNode[] = [];
  const edges: RawEdge[] = [];
  const nodeSet = new Set<string>();
  const edgeSet = new Set<string>();

  movies.forEach(movie => {
    if (!nodeSet.has(movie.id)) {
      nodes.push({
        id: movie.id,
        data: {
          type: NodeType.Movie,
          ...movie
        }
      });
      nodeSet.add(movie.id);
    }
  });

  people.forEach(person => {
    if (!nodeSet.has(person.id)) {
      nodes.push({
        id: person.id,
        data: {
          type: NodeType.Person,
          ...person
        }
      });
      nodeSet.add(person.id);
    }
  });

  genres.forEach(genre => {
    if (!nodeSet.has(genre.id)) {
      nodes.push({
        id: genre.id,
        data: {
          type: NodeType.Genre,
          ...genre
        }
      });
      nodeSet.add(genre.id);
    }
  });

  directedRelationships.forEach(relationship => {
    if (!edgeSet.has(relationship.id)) {
      edges.push({
        id: relationship.id,
        source: relationship.personId,
        target: relationship.movieId,
        data: { type: EdgeType.Directed }
      });
      edgeSet.add(relationship.id);
    }
  });

  actedInRelationships.forEach(relationship => {
    if (!edgeSet.has(relationship.id)) {
      edges.push({
        id: relationship.id,
        source: relationship.personId,
        target: relationship.movieId,
        data: { type: EdgeType.ActedIn }
      });
      edgeSet.add(relationship.id);
    }
  });

  hasGenreRelationships.forEach(relationship => {
    if (!edgeSet.has(relationship.id)) {
      edges.push({
        id: relationship.id,
        source: relationship.movieId,
        target: relationship.genreId,
        data: { type: EdgeType.HasGenre }
      });
      edgeSet.add(relationship.id);
    }
  });

  return { nodes, edges };
}
