import {
  ActedInRelationship,
  DirectedRelationship,
  Genre,
  Movie,
  Person,
  GraphEdge,
  GraphNode,
  NodeType,
  EdgeType,
  HasGenreRelationship
} from './types';

interface Row {
  movieId?: string;
  movieTitle?: string;
  movieDescription?: string;
  movieYear?: number;
  movieRuntime?: number;
  movieVotes?: number;
  directorId?: string;
  directorName?: string;
  actorId?: string;
  actorName?: string;
  genreId?: string;
  genreName?: string;
  directedId?: string;
  actedInId?: string;
  hasGenreId?: string;
}

export class Parser {
  parse(response: Row[]) {
    const movies: Movie[] = [];
    const people: Person[] = [];
    const genres: Genre[] = [];
    const directedRelationships: DirectedRelationship[] = [];
    const actedInRelationships: ActedInRelationship[] = [];
    const hasGenreRelationships: HasGenreRelationship[] = [];

    const personMap = new Map<string, Person>();
    const genreMap = new Map<string, Genre>();
    const movieMap = new Map<string, Movie>();

    const edgeSet = new Set<string>();
    const nodeSet = new Set<string>();

    const nodes: GraphNode[] = [];
    const edges: GraphEdge[] = [];

    for (const row of response) {
      const {
        movieId,
        movieTitle,
        movieDescription,
        movieYear,
        movieRuntime,
        movieVotes,
        directorId,
        directorName,
        actorId,
        actorName,
        genreId,
        genreName,
        directedId,
        actedInId,
        hasGenreId
      } = row;

      // get the MOVIE nodes from the row
      if (movieId && !nodeSet.has(movieId)) {
        const movie: Movie = {
          id: movieId,
          title: movieTitle!,
          description: movieDescription!,
          year: movieYear!,
          runtime: movieRuntime!,
          votes: movieVotes!
        };
        movies.push(movie);
        movieMap.set(movieId, movie);
        nodeSet.add(movieId);

        nodes.push({
          id: movieId,
          data: { type: NodeType.Movie, ...movie }
        });
      }

      // get the DIRECTOR nodes from the row
      if (directorId && !nodeSet.has(directorId)) {
        const director: Person = {
          id: directorId,
          name: directorName || ''
        };
        people.push(director);
        personMap.set(directorId, director);
        nodeSet.add(directorId);

        nodes.push({
          id: directorId,
          data: { type: NodeType.Person, ...director }
        });
      }

      // get the ACTOR nodes from the row
      if (actorId && !nodeSet.has(actorId)) {
        const actor: Person = {
          id: actorId,
          name: actorName || ''
        };
        people.push(actor);
        personMap.set(actorId, actor);
        nodeSet.add(actorId);

        nodes.push({
          id: actorId,
          data: { type: NodeType.Person, ...actor }
        });
      }

      // get the GENRE nodes from the row
      if (genreId && !nodeSet.has(genreId)) {
        const genre: Genre = {
          id: genreId,
          name: genreName!
        };
        genres.push(genre);
        genreMap.set(genreId, genre);
        nodeSet.add(genreId);

        nodes.push({
          id: genreId,
          data: { type: NodeType.Genre, ...genre }
        });
      }

      // get the DIRECTED relationships from the row
      if (directedId && directorId && movieId && !edgeSet.has(directedId)) {
        const directedEdge: DirectedRelationship = {
          id: directedId!,
          personId: directorId!,
          movieId: movieId
        };
        directedRelationships.push(directedEdge);
        edgeSet.add(directedId);

        edges.push({
          source: directorId,
          target: movieId,
          data: { type: EdgeType.Directed, ...directedEdge }
        });
      }

      // get the ACTED_IN relationships from the row
      if (actedInId && actorId && movieId && !edgeSet.has(actedInId)) {
        const actedIn: ActedInRelationship = {
          id: actedInId,
          personId: actorId,
          movieId: movieId
        };
        actedInRelationships.push(actedIn);
        edgeSet.add(actedInId);

        edges.push({
          source: actorId,
          target: movieId,
          data: { type: EdgeType.ActedIn, ...actedIn }
        });
      }

      // get the HAS_GENRE relationships from the row
      if (hasGenreId && genreId && movieId && !edgeSet.has(hasGenreId)) {
        const hasGenre: HasGenreRelationship = {
          id: hasGenreId,
          movieId: movieId,
          genreId: genreId
        };
        hasGenreRelationships.push(hasGenre);
        edgeSet.add(hasGenreId);

        edges.push({
          source: movieId,
          target: genreId,
          data: { type: EdgeType.HasGenre, ...hasGenre }
        });
      }
    }

    return { nodes, edges };
  }
}
